/*
* Copyright 2003(c) Theo Pavlidis
* See PixEx.h for a detailed notice on the use of the software
*/	

// Pix.cpp: implementation of the CPix class.
//
//////////////////////////////////////////////////////////////////////

//	DO NOT EDIT THIS FILE OUTSIDE p4pApp PROJECT

#include "stdafx.h"
#include <math.h>
#include "Pix.h"
#include "png.h"
#include "pngcall.h"
#include "pack4png.h"
#include "util.h"

#ifndef BATCH_P4P
#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif
#endif //BATCH_P4P

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPix::CPix()
{
	for(int i=0; i<TOTAL_P; i++)	m_Image[i] = NULL;
	m_New = OUTP1_P;
	m_Old = OUTP2_P;
}

CPix::~CPix()
{
	for(int i=0; i<TOTAL_P; i++)	if(m_Image[i] != NULL) delete[] m_Image[i];
}

int CPix::ReadPix(char *pathname)
{
	int x, y, i, j, k;
	for(i=0; i<TOTAL_P; i++)	if(m_Image[i] != NULL) delete[] m_Image[i];
	for(i=0; i<TOTAL_P; i++)	m_Image[i] = NULL;
	m_New = OUTP1_P;
	m_Old = OUTP2_P;

	png_bytep * row_pointers = NULL;
	for(i=0; i<3; i++) m_zone[i] = 0;
	png_uint_32 height, width, bpp;	// variables to match expectations of PNG library
	int status = ReadPNGfile(&row_pointers, &width, &height, &bpp, pathname);
	if(status==PNG_READ_OK) {
		m_width  = width;
		m_height = height;
		m_bpp    = bpp;
		m_Image[INPUT_P] = new unsigned char[height*width*bpp];
		switch(bpp) {
		case 1:
			for(y=0, j=0; y<(int)height; y++) {
				for(x=0; x<(int)width; x++, j++) m_Image[INPUT_P][j] = *(row_pointers[y]+x);
			}
			m_mono = 1;
			break;
		case 3:
			m_mono = 1;
			for(y=0, j=0; y<(int)height; y++) {
				for(x=0; x<(int)width; x++, j+= 3) {
					for(k=0; k<3; k++) m_Image[INPUT_P][j+k] = *(row_pointers[y]+3*x+k);
					if(m_mono==1) {
						if( (m_Image[INPUT_P][j] != m_Image[INPUT_P][j+1]) ||
							(m_Image[INPUT_P][j] != m_Image[INPUT_P][j+2]) ) m_mono = 0;
					}
				}
			}
			break;
		default:
			status = UNIMPLEMENTED_FEATURE;
		}

		m_limit[0] = m_mono==1 ? 75: 60;
		m_limit[1] = m_mono==1 ? 40: 25;
		Distribution(16);
	}

	if(row_pointers != NULL) {
		for(j=0; j<(int)height; j++) delete[] row_pointers[j];
		delete[] row_pointers;
	}
	return status;
}

int CPix::WritePix(char *pathname)
{
	if(m_Image[m_New]==NULL) return NOTHING_TO_WRITE; 
	return WritePNGfile(m_Image[m_New], m_width, m_height, m_bpp, pathname);
}

double CPix::InvokeP4P(int strategy, int ColorTableStep, int needMemory)
{
	int x, y, j, k;
	double color_err[3], total_err;	// assumes m_bpp<4
	if(needMemory > 0 && m_Image[m_New] == NULL) m_Image[m_New] = new unsigned char[m_height*m_width*m_bpp];

	//	specify nature of image for passing to packing function
	if(m_bpp>1) strategy += COLOR_P4P;
	if(m_distribution > m_limit[0]) strategy += FORCED_P4P;

	// proceed with packing
	long h[LVL];
	int vmap[LVL];
	for(k=0, total_err=0; k<m_bpp; k++) {	// for each color
		// compute histogram
		for(j=0; j<LVL; j++) h[j] = 0;
		for(y=0, j=0; y<m_height; y++) {
			for(x=0; x<m_width; x++, j += m_bpp) h[m_Image[INPUT_P][j+k]]++;
		}
		if(ColorTableStep < 1) m_zone[k] = pack4png1(h, LVL, strategy, vmap, &(color_err[k]));		// apply pack4png
		else m_zone[k] = pack4gct1(h, LVL, strategy, vmap, &(color_err[k]));
		if(m_zone[k] < 0) return -1;
		// transform image
		for(y=0, j=0; y<m_height; y++) {
			for(x=0; x<m_width; x++, j += m_bpp) m_Image[m_New][j+k] = vmap[m_Image[INPUT_P][j+k]];
		}
		total_err += color_err[k];
	}

	total_err /= m_bpp;
	return sqrt(total_err);
}

//	Histogram analysis (incl. 3D) for estimating D

static int comp(const void *e1, const void *e2)
{
	double n1 = *((double *)e1);
	double n2 = *((double *)e2);
	if(n1>n2) return 1;
	if(n1<n2) return -1;
	return 0;
}

int CPix::Hist1d(int image, int Nbins, double *result)
{
	int i, j, k;
	long hgray[LVL];
	for(j=0; j<LVL; j++) hgray[j] = 0;
	int Number = m_width*m_height*m_bpp;
	for(j=0; j<Number; j+= m_bpp) hgray[m_Image[image][j]]++;
	m_colors = 0;
	for(j=0; j<LVL; j++) if(hgray[j]>0) m_colors++;

	double *bin = new double[Nbins];	// 16*16=256
	int binsz = LVL/Nbins;
	for(i=0, j=0; i<Nbins; i++, j+=binsz) {
		bin[i]= 0;
		for(k=j; k<j+binsz; k++) bin[i] += hgray[k];
	}
	double binsum = 0;
	for(i=0; i<Nbins; i++) binsum += bin[i];
	binsum /= 100;	// to make percentages

	double threshold = 10.0;
	k = 0;
	while(k==0) {
		for(i=0; i<Nbins; i++) {
			bin[i] /= binsum;
			if(bin[i] > threshold) result[k++] = bin[i];
		}
		if(k==0) threshold /= 2;
	}

	qsort(result, k, sizeof(double), comp);
	delete[] bin;
	return k;
}

#define bin3d(A,B,C) bin3d[A*N3Dbins*N3Dbins + B*N3Dbins + C]

int CPix::Hist3d(int image, int N3Dbins, double *result)
{
	long TotalSize = N3Dbins*N3Dbins*N3Dbins;
	double *bin3d = new double[TotalSize];
	ZeroMemory(bin3d, TotalSize*sizeof(double));
	long maxColors = LVL*LVL*LVL;
	char *fullHist = new char[maxColors];
	ZeroMemory(fullHist, maxColors);
	long LVL2 = LVL*LVL;
	int ir, ig, ib;
	int oc;
	int binsz = LVL/N3Dbins;

	int Number = m_width*m_height*m_bpp;

	m_colors = 0;
	for(int j=0; j<Number; j+= m_bpp){
		oc = m_Image[image][j]*LVL2 + m_Image[image][j+1]*LVL + m_Image[image][j+2];
		if(fullHist[oc]==0) {
			m_colors++;
			fullHist[oc] = 1;
		}
		ib = m_Image[image][j]/binsz;
		ig = m_Image[image][j+1]/binsz;
		ir = m_Image[image][j+2]/binsz;
		bin3d(ir,ig,ib)++;
	}

	double sum = 0;
	for(ir=0; ir<N3Dbins; ir++) for(ig=0; ig<N3Dbins; ig++) for(ib=0; ib<N3Dbins; ib++) sum +=bin3d(ir,ig,ib);
	sum /= 100;	// to make percentages
	for(ir=0; ir<N3Dbins; ir++) for(ig=0; ig<N3Dbins; ig++) for(ib=0; ib<N3Dbins; ib++) bin3d(ir,ig,ib) /= sum;

	double threshold = m_mono==1? 10.0 : 1.5;
	int k = 0;
	while(k==0) {
		for(ir=0; ir<N3Dbins; ir++) {
			for(ig=0; ig<N3Dbins; ig++) for(ib=0; ib<N3Dbins; ib++) {
				if(bin3d(ir,ig,ib) > threshold) result[k++] = bin3d(ir,ig,ib);
			}
		}
		if(k==0) threshold /= 2;
	}
	qsort(result, k, sizeof(double), comp);
	delete[] bin3d;
	delete[] fullHist;
	return k;
}

int CPix::Distribution(int N3Dbins)
{
	CWaitCursor crunch;

	int Total3D = m_bpp==1? N3Dbins : N3Dbins*N3Dbins*N3Dbins;
	double *result = new double[Total3D];

	int kbin; 
	if(m_bpp==1) kbin = Hist1d(INPUT_P, N3Dbins, result);
	else kbin = Hist3d(INPUT_P, N3Dbins, result);

	double xd = 0;
	int low = m_mono==1? 2: 3;
	for(int i=kbin-1; i>=0 && i>=kbin-low; i--) xd += result[i];
	m_distribution = (int)(xd+0.5);

	delete[] result;
	return 1;
}
