/*
* Copyright 2003(c) Theo Pavlidis
* See PixEx.h for a detailed notice on the use of the software
*/	

// PixEx.cpp: implementation of the CPixEx class.
//
//////////////////////////////////////////////////////////////////////

//	DO NOT EDIT THIS FILE OUTSIDE p4pApp PROJECT

#include "stdafx.h"
#include <math.h>
#include "p4pApp.h"		// needed for on the spot web page display
#include "MainFrm.h"	// needed for on the spot web page display
#include "PixEx.h"
#include "png.h"
#include "pngcall.h"
#include "pack4png.h"
#include "util.h"

#ifndef BATCH_P4P
#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif
#endif //BATCH_P4P

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPixEx::CPixEx()
{
	m_levelOld = m_level = 0;
	m_colorStepOld = m_colorStep = 0;
}

CPixEx::~CPixEx()
{
}

#define InpImage (m_Image[INPUT_P])
#define InpSize  (m_PNGSize[INPUT_P])
#define InpMemDC (m_MemDC[INPUT_P])

#define OutImage (m_Image[m_New])
#define OutSize  (m_PNGSize[m_New])
#define OutMemDC (m_MemDC[m_New])

#define OldImage (m_Image[m_Old])
#define OldSize  (m_PNGSize[m_Old])
#define OldMemDC (m_MemDC[m_Old])

int CPixEx::ReadPix(char *pathname)
{
	int status = CPix::ReadPix(pathname);
	if(status != PNG_READ_OK) {
		CString msg;
		HandleReturnOfReadPNGfile(status, &msg);
		AfxMessageBox(msg);
	}
	else InpSize = FileSize(pathname);
	return status;
}

int CPixEx::WritePix(char *pathname)
{
	int status = CPix::WritePix(pathname);
	if(status != PNG_WRITE_OK) {
		CString msg;
		HandleReturnOfWritePNGfile(status, &msg);
		AfxMessageBox(msg);
	}
	return status;
}

void CPixEx::ResetOutputFiles(BOOL redraw)
{
	int i;
	for(i=1; i<TOTAL_P; i++)	if(m_Image[i] != NULL) delete[] m_Image[i];
	for(i=1; i<TOTAL_P; i++)	m_Image[i] = NULL;
	m_New = OUTP1_P;
	m_Old = OUTP2_P;
	if(redraw==TRUE) m_viewp->Invalidate();
	for(i=0; i<3; i++) m_zone[i] = 0;// to erase memory for quantization
}

// allocates space for new image
void CPixEx::Push2Old()
{
	int n = m_height*m_width*m_bpp;
	if(OutImage == NULL) OutImage = new unsigned char[n];
	else {
		if(OldImage == NULL) {
			OldImage = new unsigned char[m_height*m_width*m_bpp];
			m_PNGSize[m_Old] = m_PNGSize[m_New];
			for(int i=0; i<n; i++) OldImage[i] = OutImage[i];
			CreateMemDC(m_Old);
		}
		else {
			int tloc = m_New;
			m_New = m_Old;
			m_Old = tloc;
		}
	}
	m_levelOld = m_level;
	m_colorStepOld = m_colorStep;
	m_rmsOld = m_rms;
}

// this is used for creating the output file name - it should containt only letters or numbers
void  CPixEx::DefaultTag(char v[])
{
	if(m_level<0) {	v[0] = 'q';	v[1] = 'u';  v[2] = '\0'; }
	else { sprintf(v, "%d", m_level); }
}

void CPixEx::ShowLocations()
{
	CString tmp;
	tmp.Format("Web pages are at %s\nHelp directory is %s\nTemporary file is %s\nDefault Image directory is %s",
		WEB_DIR, HELP_DIR, TMP_FILE, IMG_DIR);
	AfxMessageBox(tmp);
}

//////////////////////////////////////////////////////////////////////////////////////////
///////////// The functions below are used only in conjuction with a display /////////////
//////////////////////////////////////////////////////////////////////////////////////////

// create copy and extract statistics
void CPixEx::FindFileSize()
{
	int status = WritePNGfile(OutImage, m_width, m_height, m_bpp, TMP_FILE);
	if(status != PNG_WRITE_OK) {
		CString msg;
		HandleReturnOfWritePNGfile(status, &msg);
		AfxMessageBox(msg);
	}
	OutSize = FileSize(TMP_FILE);
	remove(TMP_FILE); // or _unlink	// either does not seem to be working!
}

void CPixEx::InvokeP4P(int strategy, int ColorTableStep)
{
	Push2Old();
	m_level = strategy;
	m_colorStep = ColorTableStep;
	m_rms = CPix::InvokeP4P(strategy, ColorTableStep, FALSE);
	FindFileSize();
	CreateMemDC(m_New);
	m_viewp->Invalidate();
}

#include "quant.h"

void CPixEx::Quantize(int method)
{
	Push2Old();
	m_level = -1 - method;

	switch(method) {
	case MOST_SIGN_BITS:
		::MostSignBits(InpImage, m_width, m_height, m_bpp, m_zone, OutImage);
		break;
	case HIST_EQUALIZATION:
		::HistEqualization(InpImage, m_width, m_height, m_bpp, m_zone, OutImage);
		break;
	default: return;
	}
	int n = m_width*m_height*m_bpp;
	int i;
	for(i=0, m_rms=0; i<n; i++) m_rms += (InpImage[i]-OutImage[i])*(InpImage[i]-OutImage[i]);
	m_rms = sqrt(m_rms/n);

	FindFileSize();
	CreateMemDC(m_New);
	m_viewp->Invalidate();
}

int CPixEx::CreateMemDC(CView *vp, int k)	// pointer to CView
{
	m_viewp = vp;
	return CreateMemDC(k);
}

int CPixEx::CreateMemDC(int k)
{
	CClientDC dc(m_viewp);
	CBitmap bm;
	CDC *pDC = &(m_MemDC[k]);
	BYTE * img = m_Image[k];
	bm.CreateCompatibleBitmap(&dc, m_width, m_height);	// must come before memDC creation
	pDC->Detach();
	pDC->CreateCompatibleDC(&dc);
	pDC->SelectObject(&bm);
	int x, y, i;
	switch(m_bpp) {
	case 1:
		for(y=0, i=0; y<m_height; y++){
			for(x=0; x<m_width; x++, i++) pDC->SetPixel(x, y, RGB(img[i], img[i], img[i]));			
		}
		break;
	case 3:
		for(y=0, i=0; y<m_height; y++){
			for(x=0; x<m_width; x++, i += m_bpp) pDC->SetPixel(x, y, RGB(img[i], img[i+1], img[i+2]));			
		}
		break;
	default:
		return 0;
	}
	return 1;
}

static void formatN(int n, CString *sp)
{
	int kb = (n+512)/1024;
	if(n>999) sp->Format("%dK (%d,%03d)", kb, n/1000, n%1000);
	else sp->Format("%dK (%d)", kb, n);
}

static void formatC(int n, CString *sp)
{
	if(n>999) sp->Format("%d,%03d", n/1000, n%1000);
	else sp->Format("%d", n);
}

static void CreateOutText(int size, int insize, int level, int colorStep, double rms, CString *sp)
{
	CString tmp, tmpx;
	formatN(size, &tmp);
	if(level < 0) {
		switch(-level-1) {
		case CPixEx::MOST_SIGN_BITS:
			tmpx = "Quantized: ";
			break;
		case CPixEx::HIST_EQUALIZATION:
			tmpx = "Equalized: ";
			break;
		default:
			tmpx = "??: ";
		}
	}
	else {
		if(colorStep>0) tmpx = "For 6x6x6 GCT - PNG size is ";
		else tmpx.Format("Level %d,%d: ", level/10, level%10);
	}
	double pct = ((double)size*100)/insize;
	sp->Format("%s%s bytes, %d%% (e= %5.1f)", tmpx.GetBuffer(0), tmp.GetBuffer(0), (int)(pct+0.5), rms);
}

#define LEFT_MARGIN			  2
#define TEXT_GAP			  4
#define TOP_MARGIN			 50
#define TEXT_LENGTH			450

static void Layout(int w, int h, int *xp, int *yp)
{
	if(w > 2*h) {	// display below
		*yp += (h + TOP_MARGIN);
	}
	else {
		if(w < TEXT_LENGTH) *xp = TEXT_LENGTH;
		else *xp += (LEFT_MARGIN+w);
	}
}

void CPixEx::TrafficSign(CDC *pDC, int xe, int ye)
{
	CBrush redBrush(RGB(255, 0, 0));
	CBrush greenBrush(RGB(0, 255, 0));
	CBrush amberBrush(RGB(255, 200, 0));
	if(m_distribution > m_limit[0]) pDC->SelectObject(&greenBrush);
	else {
		if(m_distribution > m_limit[1]) pDC->SelectObject(&amberBrush);
		else pDC->SelectObject(&redBrush);
	}
	int bound = (m_mono==1)? 40: 4000;
	if(m_colors > bound) pDC->Ellipse(xe, ye-2, xe+20, ye+18);
	else pDC->Ellipse(xe, ye+3, xe+10, ye+13);
	pDC->SelectStockObject(NULL_BRUSH);
}

void CPixEx::DrawOnePix(CDC *pDC, int image, int xoff, int yoff, CString *msgp)
{
	CSize sz = pDC->GetTextExtent(*msgp);
	pDC->TextOut(xoff, yoff - sz.cy-TEXT_GAP, msgp->GetBuffer(0));

	if(image==INPUT_P) TrafficSign(pDC, xoff+sz.cx+TEXT_GAP, yoff - sz.cy-TEXT_GAP);

	pDC->BitBlt(xoff, yoff, m_width, m_height, &(m_MemDC[image]), 0, 0, SRCCOPY);
	pDC->Rectangle(xoff, yoff, xoff+m_width, yoff+m_height);
}

void CPixEx::DrawPix(CDC* pDC)
{
	CString msg, tmp, tmpx;
	int xoff = LEFT_MARGIN, yoff = TOP_MARGIN;
	pDC->SelectStockObject(BLACK_PEN);
	pDC->SelectStockObject(NULL_BRUSH);
	//	Display left panel - old image, or if there is none, input image
	if(OldImage != NULL) {
		CreateOutText(OldSize, InpSize, m_levelOld, m_colorStepOld, m_rmsOld, &msg);
		DrawOnePix(pDC, m_Old, xoff, yoff, &msg);
	}
	else {
		if(InpImage != NULL) {
			formatN(InpSize, &tmp);
			formatC(m_colors, &tmpx);
			msg.Format("%dx%dx%d; %s bytes, %s colors, D=%d",
				m_width, m_height, m_bpp, tmp.GetBuffer(0), tmpx.GetBuffer(0), m_distribution);
			DrawOnePix(pDC, INPUT_P, xoff, yoff, &msg);
		}
	}
	//	Display right (or below) panel - new image goes always here
	if(OutImage != NULL) {
		Layout(m_width, m_height, &xoff, &yoff);
		CreateOutText(OutSize, InpSize, m_level, m_colorStep, m_rms, &msg);
		DrawOnePix(pDC, m_New, xoff, yoff, &msg);
	}
}

#include <direct.h>

#define IMGBLOCK(A,B) iblock[A].Format(ImageTemplate,	\
		Dirname.GetBuffer(0), B.GetBuffer(0), Dirname.GetBuffer(0), B.GetBuffer(0),	\
		(int)(m_width*scale), (int)(m_height*scale))

#define TOO_MANY_ZONES	12	// six peaks
#define TOO_OFTEN		 4	// out of 24 possible for color, or 8 possible for gray

void CPixEx::Web(CString FileTitle)
{
	CWaitCursor crunch;
	if(_chdir(WEB_DIR) != 0) {
		AfxMessageBox("Cannot Change Directory");
		return;
	}
	// construct web name from file title
	CString Dirname, WebTitle, OutFile, tmp, tmpx;
	int lastDot = FileTitle.ReverseFind('.');
	ASSERT(lastDot>0);
	WebTitle = FileTitle.Left(lastDot);
	//	Create Directory in web directory
	Dirname.Format("%s_Files", WebTitle.GetBuffer(0));
	_mkdir(Dirname.GetBuffer(0));
	//	Create Web File name
	WebTitle += ".htm";

		// estimate display scale
	double scale;
	// Assume a 600 by 1000 page
	double vscale, hscale;
	if(m_height > m_width/2) {
		hscale = 200.0/m_width;	vscale = 300.0/m_height;
	}
	else {	// pile vertically
		hscale = 500.0/m_width;	vscale = 100.0/m_height;	// 9 rows plus
	}
	scale = hscale < vscale? hscale: vscale;
	if(scale > 1) scale = 1;

	FILE *fp = fopen(WebTitle.GetBuffer(0), "w");
	if(fp==NULL) return;	// message ?
	//	create head of page
	fputs("<html><head>\n", fp);
	tmp.Format("testing pack4png for %s", FileTitle);
	fputs("<title>", fp); fputs(tmp.GetBuffer(0), fp); fputs("</title>\n", fp);
	fputs("<style>\n<!--\ntable\n{\nfont-family:Arial, Helvetica;\nfont-size:10pt;\n}\n", fp);
	fputs("td\n{\ntext-align:center;\n}\n", fp);
	fputs("-->\n</style>\n", fp);
	fputs("</head>\n<body>\n", fp);
	// head is over - we start with body
	fputs("<table><tr><td>\n", fp);
	tmp.MakeUpper();
	fputs("<H3>", fp); fputs(tmp.GetBuffer(0), fp); fputs("</H3>\n", fp);

	// create image and text blocks for later layout
	int status;
	double pct, pctmin = 100;
	CString iblock[9], tblock[9];
	char ImageTemplate[] = "<td><a href=\"%s/%s\"><img border=\"1\" src=\"%s/%s\" width=\"%d\" height=\"%d\"></a>\n";

	IMGBLOCK(0, FileTitle);
	tmp.Format("%s/%s", Dirname, FileTitle);	// create copy of original
	status = WritePNGfile(InpImage, m_width, m_height, m_bpp, tmp.GetBuffer(0));
	ASSERT(status==PNG_WRITE_OK);
	formatN(InpSize, &tmp);
	tblock[0].Format("<td>Original: %s", tmp.GetBuffer(0));

	int st[8] = WEB_TEST_VALUES;
	int i, j, mxzone;
	mxzone = 0;
	for(i=0; i<8; i++) {
		ResetOutputFiles(FALSE);	// to avoid problems with cyclical buffer
		m_rms = CPix::InvokeP4P(st[i], 0, TRUE);
		OutFile.Format("%s/%d_%s", Dirname, st[i], FileTitle);
		status = WritePNGfile(OutImage, m_width, m_height, m_bpp, OutFile.GetBuffer(0));
		ASSERT(status==PNG_WRITE_OK);
		OutSize = FileSize(OutFile.GetBuffer(0));
		ASSERT(OutSize >=0);
		tmp.Format("%d_%s", st[i], FileTitle);
		IMGBLOCK(i+1, tmp);
		pct = ((double)OutSize*100)/((double)InpSize);
		if(pct < pctmin) pctmin = pct;
		if(m_bpp==3) tmpx.Format("/%d/%d/%d/ e=%d", m_zone[0], m_zone[1], m_zone[2], (int)(m_rms+0.5));
		else tmpx.Format("/%d/ e=%d", m_zone[0], (int)(m_rms+0.5));
		for(j=0; j<m_bpp; j++) if(m_zone[j] > TOO_MANY_ZONES) mxzone++;
		formatN(OutSize, &tmp);
		tblock[i+1].Format("<td>%d,%d: %s (%3.1f%%) %s",
			st[i]/10, st[i]%10, tmp.GetBuffer(0), pct, tmpx.GetBuffer(0));
	}
	ResetOutputFiles(FALSE);	// to avoid problems with cyclical buffer

	// original compression compared to bitmap
	double ocomp = (double)InpSize/(m_width*m_height*m_bpp);
	tmp.Format("Original PNG size vs bitmap: %d%%. - ", (int)(ocomp*100+0.5));
	fputs(tmp.GetBuffer(0), fp);
	if(ocomp < 0.10) {
		fputs("Image is <b>not</b> likely to benefit from pack4png<br>\n", fp);
	}
	tmp.Format("Best compression after pack4png: %3.1f%%<br>\n", ocomp*pctmin);
	fputs(tmp.GetBuffer(0), fp);
	formatC(m_colors, &tmpx);
	tmp.Format("%s colors. ", tmpx.GetBuffer(0));
	fputs(tmp.GetBuffer(0), fp);

	if(m_bpp==3) tmp.Format("Percentage of pixels in three most frequent color ranges <b>D = %d%%</b><br>\n", m_distribution);
	else tmp.Format("Percentage of pixels in two most frequent color ranges <b>D = %d%%</b><br>\n", m_distribution);
	fputs(tmp.GetBuffer(0), fp);

	//if(mxzone > TOO_OFTEN) fputs("<b>CAUTION: Results suggest that image may not be suitable for PNG</b><br>", fp);
	fputs("Listed under each image: Space in bytes, percentages vs original PNG, number of zones /R/G/B/ rms error<br>", fp);
	if(scale != 1) {
		tmp.Format("Display Scale: %d%% - Click on an image to see full size<br>", (int)(scale*100+0.5));
		fputs(tmp.GetBuffer(0), fp);
	}
	else fputs("<br>", fp);

	// display blocks
	fputs("\n<table cellpadding=5><tr>", fp);
	if(m_height > m_width/2) {
		for(i=0 ; i<9; i += 3) {
			fputs(iblock[i], fp);	fputs(iblock[i+1], fp);	fputs(iblock[i+2], fp);
			fputs("</tr>\n<tr>", fp);
			fputs(tblock[i], fp);	fputs(tblock[i+1], fp);	fputs(tblock[i+2], fp);
			if(i<8) fputs("</tr>\n<tr>", fp);
		}
	}
	else {
		for(i=0; i<9; i++) {
			fputs(iblock[i], fp); fputs(tblock[i], fp); fputs("</tr>\n<tr>", fp);
		}
	}
	fputs("</tr></table>", fp);

	fputs("</tr></table>\n", fp);
	fputs("</body></html>", fp);
	fclose(fp);
	// the code below is for displaying the web page immediately
	char buf[128];

	sprintf(buf,"%s\\%s", ".", WebTitle.GetBuffer(0));
	if( (int)ShellExecute(((CMainFrame *)AfxGetMainWnd())->m_hWnd, NULL, buf, NULL, NULL, SW_SHOWNORMAL) <33)
		AfxMessageBox("Failed to find Web Page just created");	
}
