/*
* Copyright 2003(c) Theo Pavlidis
* See PixEx.h for a detailed notice on the use of the software
*/
	
//	Invocation of Reading  and Writing functions of the PNG library
//	Consolidated 3/20/03

#include "stdafx.h"
#include "pngcall.h"

#ifndef BATCH_P4P
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
#endif //BATCH_P4P

// PGN calls

//	READING

void read_row_callback(png_structp png_ptr, png_uint_32 row_number, int pass)
{
#ifndef BATCH_P4P
	if(row_number%256 == 0) TRACE("row %d (%d) read\n", row_number, pass);
#endif
}

int ReadPNGfile(png_bytep ** prow_pointers, png_uint_32 *pwidth, png_uint_32 *pheight, png_uint_32 *pbytes_per_pixel, char *name)
{
	FILE *fp = fopen(name, "rb");
	if(fp==NULL) return FAILED_TO_OPEN_FILE;
	// Check if file PNG
	unsigned char header[8];
	int number = 4;	// bytes to read to etst if file is PNG
	fread(header, 1, number, fp);
    int is_png = !png_sig_cmp(header, 0, number);
    if (is_png==0) {
		fclose(fp);
		return NOT_A_PNG_FILE;
	}
	// Allocate memory for structures
	png_structp read_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, png_voidp_NULL,
		png_error_ptr_NULL, png_error_ptr_NULL);
	if(read_ptr==NULL) return NO_MEMORY_TO_ALLOCATE;

	png_infop read_info_ptr = png_create_info_struct(read_ptr);
	if(read_info_ptr==NULL) {
		png_destroy_read_struct(&read_ptr, (png_infopp)NULL, (png_infopp)NULL);
		return NO_MEMORY_TO_ALLOCATE;
	}
	png_infop end_info_ptr  = png_create_info_struct(read_ptr);
	if(end_info_ptr==NULL) {
		png_destroy_read_struct(&read_ptr, &read_info_ptr, (png_infopp)NULL);
		return NO_MEMORY_TO_ALLOCATE;
	}

	// indices, parameters, and variables to allocate memory to later on (needed in setjmp())
	int i;
	unsigned long width, height;
	int bit_depth, color_type, interlace_type, compression_type, filter_type;
	//png_bytep * row_pointers = NULL;
	// emergency return from deep in PNG
	if (setjmp(png_jmpbuf(read_ptr))) {
		png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
		fclose(fp);
		return ERROR_FROM_DEEP_PNG;
	}
	// getting ready to read
	png_init_io(read_ptr, fp);
	png_set_sig_bytes(read_ptr, number);	// let it know about bytes already read to test file type
	png_set_read_status_fn(read_ptr, read_row_callback);

	//	read header
	png_read_info(read_ptr, read_info_ptr);
	//extract data from it
	int Header_status = png_get_IHDR(read_ptr, read_info_ptr, &width, &height,
       &bit_depth, &color_type, &interlace_type,
       &compression_type, &filter_type);

	if(color_type != PNG_COLOR_TYPE_GRAY && color_type != PNG_COLOR_TYPE_RGB) {
#ifdef PLAY_IT_SAFE
		png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
		fclose(fp);
		return CANNOT_HANDLE_PALETTE;
#endif
	}

	int channels = png_get_channels(read_ptr, read_info_ptr);
	//int rowbytes = png_get_rowbytes(read_ptr, read_info_ptr);
	//ASSERT(rowbytes==(channels*(int)width));	// this may not work if width is an odd number

	// read palette 
	png_colorp palette;
	int num_palette;
	int Pal_status = png_get_PLTE(read_ptr, read_info_ptr, &palette, &num_palette);
	if(Pal_status != 0) {	// we never reach this now
		png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
		fclose(fp);
		return CANNOT_HANDLE_PALETTE;
	}

	int srgb_intent;
	if(png_get_sRGB(read_ptr, read_info_ptr, &srgb_intent) > 0) {
		// make use of information - we do not have any of that now
	}
	//	look for text
    png_textp text_ptr;
	int num_text;
	if (png_get_text(read_ptr, read_info_ptr, &text_ptr, &num_text) > 0) {
		// make use of information - we ignore it for now
	}

	*prow_pointers = new png_bytep[height];
	int row_buf_size = png_get_rowbytes(read_ptr, read_info_ptr);
	for(i=0; i<(int)height; i++) (*prow_pointers)[i] = new png_byte[row_buf_size];

	png_read_image(read_ptr, *prow_pointers);
	*pheight = height;
	*pwidth  = width;
//	*pbytes_per_pixel = color_type==PNG_COLOR_TYPE_GRAY ? 1 : 3;	// a bit careless here INDEED
	*pbytes_per_pixel = channels;	// that seems better (3/27/03)

	png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
	fclose(fp);
	return PNG_READ_OK;
}

#ifndef BATCH_P4P
void HandleReturnOfReadPNGfile(int n, CString *msgp)
{
	CString tmp;
	*msgp = "PNG: ";
	switch(n) {
	case CANNOT_HANDLE_PALETTE: *msgp += "File has Palette\nCode for that feature is not implemented"; break;
	case UNIMPLEMENTED_FEATURE:	*msgp += "File read succesfully\nbut it has unimplemented features"; break;
	case FAILED_TO_OPEN_FILE:	*msgp += "Failed to find file";	break;
	case NOT_A_PNG_FILE:		*msgp += "File is not PNG";	break;
	case NO_MEMORY_TO_ALLOCATE:	*msgp += "Failed to allocate memory\nfor PNG structures"; break;
	case ERROR_FROM_DEEP_PNG:	*msgp += "Error from inside PNG routines"; break;
	case PNG_READ_OK:			*msgp += "Normal Completion"; break;
	default:					msgp->Format("PNG: ??? %d", n);
	}
}
#endif

//	WRITING

void write_row_callback(png_structp png_ptr, png_uint_32 row_number, int pass)
{
#ifndef BATCH_P4P
	if(row_number%256 == 0) TRACE("row %d (%d) written\n", row_number, pass);
#endif
}

int WritePNGfile(png_byte image[], png_uint_32 width, png_uint_32 height, png_uint_32 bytes_per_pixel, char *name)
{
	unsigned int k;
	int bit_depth	= 8;

	FILE *fp = fopen(name, "wb");
	if(fp==NULL) return FAILED_TO_CREATE_FILE;

	png_structp write_ptr  = png_create_write_struct(PNG_LIBPNG_VER_STRING, png_voidp_NULL,
	  png_error_ptr_NULL, png_error_ptr_NULL);
	if(write_ptr==NULL) return NO_MEMORY_TO_ALLOCATE;

    png_infop info_ptr = png_create_info_struct(write_ptr);
    if (info_ptr==NULL)     {
		png_destroy_write_struct(&write_ptr, (png_infopp)NULL);
		return NO_MEMORY_TO_ALLOCATE;
    }

	// we allocate here rather than before output, so memory can be freed in case of internal PNG errors
	png_bytep *row_pointers = new png_bytep[height];
	if(row_pointers==NULL) {
		png_destroy_write_struct(&write_ptr, (png_infopp)NULL);
		return NO_MEMORY_TO_ALLOCATE;
	}
	for(k=0; k<height; k++) row_pointers[k] = image + k*width*bytes_per_pixel;

	// emergency return from deep in PNG
    if (setjmp(png_jmpbuf(write_ptr))) {
		png_destroy_write_struct(&write_ptr, &info_ptr);
		delete[] row_pointers;
		fclose(fp);
		return ERROR_FROM_DEEP_PNG;
    }

	png_init_io(write_ptr, fp);
	png_set_write_status_fn(write_ptr, write_row_callback);
	// initialize information structure
	int color_type = bytes_per_pixel==1? PNG_COLOR_TYPE_GRAY: PNG_COLOR_TYPE_RGB;
	png_set_IHDR(write_ptr, info_ptr, width, height, bit_depth,
		color_type,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

#ifdef BEING_FUSSY
	png_set_sRGB_gAMA_and_cHRM(write_ptr, info_ptr, PNG_sRGB_INTENT_SATURATION);
#endif

	png_text text_ptr[1];
	text_ptr[0].key  = "Title";
	text_ptr[0].text = "PACK4PNG TEST";
	text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
#ifdef PNG_iTXt_SUPPORTED
	text_ptr[0].lang = NULL;
#endif
	png_set_text(write_ptr, info_ptr, text_ptr, 1);

	//	perform output
	png_write_info(write_ptr, info_ptr);
	png_write_image(write_ptr, row_pointers);
	png_write_end(write_ptr, NULL);

	png_destroy_write_struct(&write_ptr, &info_ptr);
	delete[] row_pointers;
	fclose(fp);
	return PNG_WRITE_OK;
}

#ifndef BATCH_P4P
void HandleReturnOfWritePNGfile(int n, CString *msgp)
{
	*msgp = "PNG: ";
	switch(n) {
	case NOTHING_TO_WRITE:		*msgp += "No data to write"; break;
	case FAILED_TO_CREATE_FILE: *msgp += "Cannot create PNG file"; break;
	case NO_MEMORY_TO_ALLOCATE: *msgp += "Memory allocation failed"; break;
	case ERROR_FROM_DEEP_PNG:	*msgp += "Error from inside PNG routines"; break;
	case PNG_WRITE_OK:			*msgp += "Normal Completion"; break;
	default: msgp->Format("PNG: ??? %d", n);
	}
}
#endif